/* Author:  Yotam Medini  yotam.medini@gmail.com -- Created: 2011/March/26 */

#include "phash.h"
#include <stdlib.h>
#include <string.h>
#include "strhash.h"


typedef unsigned long  ul_t;


static ul_t  prime_larger_than(ul_t x)
{
    /* Table taken from C++ STL's hashtable.h */
    static const ul_t primes[] =
      {
        53ul,         97ul,         193ul,       389ul,       769ul,
        1543ul,       3079ul,       6151ul,      12289ul,     24593ul,
        49157ul,      98317ul,      196613ul,    393241ul,    786433ul,
        1572869ul,    3145739ul,    6291469ul,   12582917ul,  25165843ul,
        50331653ul,   100663319ul,  201326611ul, 402653189ul, 805306457ul,
        1610612741ul, 3221225473ul, 4294967291ul
      };

    unsigned n_primes = sizeof(primes) / sizeof(primes[0]);
    unsigned i;

    for (i = 0; (i + 1 < n_primes) && (primes[i] <= x); i++);
    return primes[i];
}


static void phash_clear_free_persons(phash_table_t *phtbl, int do_free_persons)
{
    ul_t hi;

    for (hi = 0; hi != phtbl->hash_size; hi++)
    {
        person_node_t  *pnode, *pnext;
        for (pnode = phtbl->entries[hi];  pnode;  pnode = pnext)
        {
            pnext = pnode->next;
            if (do_free_persons)
            {
                person_free(pnode->person);
            }
            free(pnode);
        }
    }
    free(phtbl->entries);
    phash_init(phtbl);
}


static void phash_enlarge(phash_table_t *phtbl)
{
    phash_table_t  new_tbl;
    ul_t           hi;
    person_node_t  *pnode;

    new_tbl.n_persons = 0;
    new_tbl.hash_size = prime_larger_than(phtbl->n_persons);
    new_tbl.entries = calloc(new_tbl.hash_size, sizeof(person_node_t *));

    for (hi = 0;  hi < phtbl->hash_size;  hi++)
    {
        for (pnode = phtbl->entries[hi];  pnode;  pnode = pnode->next)
        {
            phash_insert(&new_tbl, pnode->person);
        }
    }

    /* Clear the current table, but do NOT free the persons! */
    phash_clear_free_persons(phtbl, 0);
    
    /* Move the entries from the new table to ours */
    phtbl->entries = new_tbl.entries;
    phtbl->hash_size = new_tbl.hash_size;
    phtbl->n_persons = new_tbl.n_persons;
}


static person_node_t *phash_node_find(
    phash_table_t *phtbl,
    char          *name,
    ul_t          *hi_p)
{
    person_node_t  *pnode = 0;

    if (phtbl->hash_size > 0)
    {
        ul_t  hi = str_hash(name) % phtbl->hash_size;
        /* Check if p is already there */
        for (pnode = phtbl->entries[hi]; 
             pnode && strcmp(pnode->person->name, name);
             pnode = pnode->next);
        if (hi_p)
        { 
            *hi_p = hi;
        }
    }
    return pnode;
}


void phash_init(phash_table_t *phtbl)
{
    phtbl->entries = NULL;
    phtbl->hash_size = 0;
    phtbl->n_persons = 0;
}


int phash_insert(phash_table_t *phtbl, person_t *p)
{
    int  inserted = 0;
    ul_t  hi;

    if (phtbl->hash_size <= phtbl->n_persons)
    {
        /* Enlarge the table entries. Will alwyas happen on 1st insertion! */
        phash_enlarge(phtbl);
    }
    if (phash_node_find(phtbl, p->name, &hi) == NULL)
    {
        person_node_t  *pnext =  phtbl->entries[hi];
        person_node_t  *pnode = malloc(sizeof(person_node_t));
        pnode->person = p;
        pnode->prev = NULL;
        pnode->next = pnext;
        if (pnext) 
        {
            pnext->prev = pnode;
        }
        phtbl->entries[hi] = pnode;

        inserted = 1;
        phtbl->n_persons++;
    }

    return inserted;
}


person_t  *phash_find(phash_table_t *phtbl, char *name)
{
    person_node_t  *pnode = phash_node_find(phtbl, name, NULL);
    person_t       *p = (pnode ? pnode->person : NULL);
    return p;
}


int phash_delete(phash_table_t *phtbl, char *name)
{
    ul_t hi;
    int deleted = 0;

    person_node_t  *pnode = phash_node_find(phtbl, name, &hi);
    if (pnode)
    {
        person_t *p = pnode->person;
        person_node_t *pnext = pnode->next;
        if (phtbl->entries[hi] == pnode)
        {
            phtbl->entries[hi] = pnext;
        }
        else /* We must have prev! */
        {
            pnode->prev->next = pnext;
        }
        if (pnext)
        {
            pnext->prev = pnode->prev;
        }
        person_free(p);
        free(pnode);

        deleted = 1;
        phtbl->n_persons--;
        
    }
    return deleted;
}


void phash_clear(phash_table_t *phtbl)
{
    phash_clear_free_persons(phtbl, 1);
}
